@@ -1,18 +1,25 @@
 /*
  * picasm -- expr.c
  *
+ * Copyright 1995-2000 Timo Rossi, <trossi@iki.fi>
+ * See the file LICENSE for license terms.
+ *
  * expression parser
  *
- * Timo Rossi <trossi@iki.fi>
- * 
  */
 
 #include <stdio.h>
 #include <string.h>
 
 #include "picasm.h"
+#include "token.h"
+#include "symtab.h"
 
-int expr_error; /* expression error flag */
+/*forward*/
+static long expr_main(void);
+
+
+static int expr_error; /* expression error flag */
 
 /*
  * expression parser, bottom level
@@ -22,7 +29,7 @@ expr_element(void)
 {
   long val, tval;
   struct symbol *sym;
-  static unsigned char strbuf[256];
+  static char strbuf[256];
   int symtype;
 
   if(expr_error)
@@ -33,8 +40,8 @@ expr_element(void)
       get_token();
       tval = 0;
       while(!expr_error && token_type != TOK_RIGHTBRAK) {
-	val = get_expression();
-	if(val < 0 || val >= EXPR_NBITS) {
+	val = expr_main();
+	if(!expr_error && (val < 0 || val >= EXPR_NBITS)) {
 	  error(1, "bit number out of range");
 	  expr_error = 1;
 	}
@@ -46,7 +53,7 @@ expr_element(void)
 
     case TOK_LEFTPAR:
       get_token();
-      val = get_expression();
+      val = expr_main();
       if(!expr_error && token_type != TOK_RIGHTPAR) {
 	error(1, "')' expected");
 	expr_error = 1;
@@ -187,11 +194,11 @@ expr_element(void)
       }
 
       get_token();
-      val = get_expression();
+      val = expr_main();
       if(val < 0 || val >= (long)strlen(strbuf))
 	val = -1;
       else
-	val = strbuf[val];
+	val = (unsigned char)(strbuf[val]);
 
       if(token_type != TOK_RIGHTPAR) {
 	error(1, "')' expected");
@@ -384,13 +391,11 @@ expr_add_sub(void)
  * must be used instead)
  */
 long
-get_expression(void)
+expr_main(void)
 {
   int op;
   long val1, val2;
 
-  expr_error = 0;
-
   val1 = expr_add_sub();
   while(!expr_error && ((op = token_type) == TOK_EQ ||
 		       op == TOK_NOT_EQ ||
@@ -424,5 +429,22 @@ get_expression(void)
 	break;
     }
   }
+
   return val1;
 }
+
+int
+get_expression(long *val_p)
+{
+  long val;
+
+  expr_error = 0;
+
+  val = expr_main();
+
+  if(expr_error)
+    return FAIL;
+
+  *val_p = val;
+  return OK;
+}
