@@ -1,13 +1,16 @@
 /*
  * picasm -- pic12bit.c
  *
- * Timo Rossi <trossi@iki.fi>
+ * Copyright 1995-2000 Timo Rossi, <trossi@iki.fi>
+ * See the file LICENSE for license terms.
  * 
  */
 
 #include <stdio.h>
 
 #include "picasm.h"
+#include "token.h"
+#include "symtab.h"
 
 /*
  * Assemble 12-bit PIC code
@@ -19,18 +22,22 @@ assemble_12bit_mnemonic(int op)
   long val;
   struct symbol *sym;
   char *cp;
-  int t, symtype;
+  int symtype;
+  long t;
 
   switch(op) {
     case KW_ANDLW:
     case KW_IORLW:
     case KW_XORLW:
-      val = get_expression();
-      if(expr_error)
+      if(get_expression(&val) != OK)
 	return FAIL;
+
       if(val < -0x80 || val > 0xff)
 	error(1, "8-bit literal out of range");
-      val &= 0xff; /* this assumes 2-complement negative numbers */
+
+      if(val < 0)
+	val = 0x100 - val;
+
       switch(op) {
         case KW_ANDLW: gen_code(0xe00 | val); break;
 	case KW_IORLW: gen_code(0xd00 | val); break;
@@ -57,17 +64,21 @@ assemble_12bit_mnemonic(int op)
     case KW_RLF:
     case KW_RRF:
     case KW_SWAPF:
-      val = get_expression();
-      if(expr_error)
+
+      if(get_expression(&val) != OK)
 	return FAIL;
+
       if(val < 0 || val > 0x1f)
 	error(0, "Register file address out of range");
+
       t = 1;
       if(token_type == TOK_COMMA) {
 	get_token();
-	t = get_expression();
-	if(expr_error)
+	if(get_expression(&t) != OK)
 	  return FAIL;
+
+	if(warnlevel > 0 && t != 0 && t != 1)
+	  warning("Destination specifier not 0 or 1");
       }	else {
 	if(warnlevel > 0)
 	  warning("Destination speficier omitted");
@@ -93,9 +104,9 @@ assemble_12bit_mnemonic(int op)
 
     case KW_CLRF:
     case KW_MOVWF:
-      val = get_expression();
-      if(expr_error)
+      if(get_expression(&val) != OK)
 	return FAIL;
+
       if(val < 0 || val > 0x1f)
 	error(0, "Register file address out of range");
       switch(op) {
@@ -108,19 +119,20 @@ assemble_12bit_mnemonic(int op)
     case KW_BSF:
     case KW_BTFSC:
     case KW_BTFSS:
-      val = get_expression();
-      if(expr_error)
+      if(get_expression(&val) != OK)
 	return FAIL;
+
       if(val < 0 || val > 0x1f)
 	error(0, "Register file address out of range");
+
       if(token_type != TOK_COMMA) {
 	error(1, "',' expected");
 	return FAIL;
       }
       get_token();
-      t = get_expression();
-      if(expr_error)
+      if(get_expression(&t) != OK)
 	return FAIL;
+
       if(t < 0 || t > 7) {
 	error(0, "Bit number out of range");
       }
@@ -161,8 +173,7 @@ assemble_12bit_mnemonic(int op)
 	}
       }
 
-      val = get_expression();
-      if(expr_error)
+      if(get_expression(&val) != OK)
 	return FAIL;
 
       if(val < 0 || val >= prog_mem_size)
@@ -181,17 +192,22 @@ gen_goto_call:
       break;
 
     case KW_TRIS:
-      t = get_expression();
-      if(expr_error)
+      if(get_expression(&t) != OK)
 	return FAIL;
+
       if(t != 5 && t != 6 && t != 7)
 	error(0, "Invalid register address for TRIS");
+
       gen_code(0x000 | t);
       break;
 
 /*
  * RETLW allows multiple parameters/strings, for generating lookup tables
+ *
+ * DT is a Microchip-compatible directive that generates retlw instructions
+ * it is handled exactly the same as the actual retlw mnemonic here
  */
+    case KW_DT:
     case KW_RETLW:
       for(;;) {
 	if(token_type == TOK_STRCONST) {
