@@ -1,13 +1,16 @@
 /*
  * picasm -- pic14bit.c
  *
- * Timo Rossi <trossi@iki.fi>
+ * Copyright 1995-2000 Timo Rossi, <trossi@iki.fi>
+ * See the file LICENSE for license terms.
  * 
  */
 
 #include <stdio.h>
 
 #include "picasm.h"
+#include "token.h"
+#include "symtab.h"
 
 /*
  * Assemble 14-bit PIC code
@@ -18,19 +21,23 @@ assemble_14bit_mnemonic(int op)
   long val;
   struct symbol *sym;
   char *cp;
-  int t, symtype;
+  int symtype;
+  long t;
 
   switch(op) {
     case KW_SUBLW:
     case KW_ANDLW:
     case KW_IORLW:
     case KW_XORLW:
-      val = get_expression();
-      if(expr_error)
+      if(get_expression(&val) != OK)
 	return FAIL;
+
       if(val < -0x80 || val > 0xff)
 	error(0, "8-bit literal out of range");
-      val &= 0xff; /* this assumes 2-complement negative numbers */
+
+      if(val < 0)
+	val = 0x100 - val;
+
       switch(op) {
         case KW_SUBLW: gen_code(0x3c00 | val); break;
 	case KW_ANDLW: gen_code(0x3900 | val); break;
@@ -63,17 +70,20 @@ assemble_14bit_mnemonic(int op)
     case KW_RLF:
     case KW_RRF:
     case KW_SWAPF:
-      val = get_expression();
-      if(expr_error)
+      if(get_expression(&val) != OK)
 	return FAIL;
+
       if(val < 0 || val > 0x7f)
 	error(0, "Register file address out of range");
+
       t = 1;
       if(token_type == TOK_COMMA) {
 	get_token();
-	t = get_expression();
-	if(expr_error)
+	if(get_expression(&t) != OK)
 	  return FAIL;
+
+	if(warnlevel > 0 && t != 0 && t != 1)
+	  warning("Destination specifier not 0 or 1");
       }	else {
 	if(warnlevel > 0)
 	  warning("Destination speficier omitted");
@@ -100,11 +110,12 @@ assemble_14bit_mnemonic(int op)
 
     case KW_CLRF:
     case KW_MOVWF:
-      val = get_expression();
-      if(expr_error)
+      if(get_expression(&val) != OK)
 	return FAIL;
+
       if(val < 0 || val > 0x7f)
 	error(0, "Register file address out of range");
+
       switch(op) {
         case KW_CLRF:  gen_code(0x0180 | val); break;
 	case KW_MOVWF: gen_code(0x0080 | val); break;
@@ -115,19 +126,20 @@ assemble_14bit_mnemonic(int op)
     case KW_BSF:
     case KW_BTFSC:
     case KW_BTFSS:
-      val = get_expression();
-      if(expr_error)
+      if(get_expression(&val) != OK)
 	return FAIL;
+
       if(val < 0 || val > 0x7f)
 	error(0, "Register file address out of range");
+
       if(token_type != TOK_COMMA) {
 	error(1, "',' expected");
 	return FAIL;
       }
       get_token();
-      t = get_expression();
-      if(expr_error)
+      if(get_expression(&t) != OK)
 	return FAIL;
+
       if(t < 0 || t > 7) {
 	error(0, "Bit number out of range");
       }
@@ -167,11 +179,12 @@ assemble_14bit_mnemonic(int op)
 	}
       }
 
-      val = get_expression();
-      if(expr_error)
+      if(get_expression(&val) != OK)
 	return FAIL;
+
       if(val < 0 || val >= prog_mem_size)
 	error(0, "GOTO/CALL address out of range");
+
 gen_goto_call:
       val &= 0x7ff;
       switch(op) {
@@ -183,11 +196,12 @@ gen_goto_call:
 	break;
 
     case KW_TRIS:
-      t = get_expression();
-      if(expr_error)
+      if(get_expression(&t) != OK)
 	return FAIL;
+
       if(t < 5 || t > 7)
 	error(0, "Invalid register address for TRIS");
+
       gen_code(0x0060 | t);
       if(warnlevel > 1)
 	warning("TRIS instruction not recommended");
@@ -195,7 +209,11 @@ gen_goto_call:
 
 /*
  * RETLW allows multiple parameters/strings, for generating lookup tables
+ *
+ * DT is a Microchip-compatible directive that generates retlw instructions
+ * it is handled exactly the same as the actual retlw mnemonic here
  */
+    case KW_DT:
     case KW_RETLW:
       for(;;) {
 	if(token_type == TOK_STRCONST) {
@@ -219,7 +237,7 @@ gen_goto_call:
       break;
 
     case KW_CLRW:
-      gen_code(0x0100);
+      gen_code(0x0103); /* read STATUS instead of INDF */
       break;
 
     case KW_OPTION:
