@@ -1,17 +1,10 @@
 /*
  * picasm.c
  *
- * Freely distributable for non-commercial use (basically,
- * don't sell this program without my permission. You can
- * use it for developing commercial PIC applications, but
- * of course I am not responsible for any damages caused by
- * this assembler generating bad code or anything like that)
+ * Copyright 1995-2000 Timo Rossi, <trossi@iki.fi>
+ * See the file LICENSE for license terms.
  *
- * Copyright 1995-1998 by Timo Rossi
- * See the file picasm.doc for more information
- * 
- *   email: trossi@iki.fi
- *   www: http://www.iki.fi/trossi/pic/
+ * http://www.iki.fi/trossi/pic/
  *
  */
 
@@ -21,8 +14,11 @@
 #include <stdarg.h>
 #include <ctype.h>
 #include <time.h>
+#include <errno.h>
 
 #include "picasm.h"
+#include "token.h"
+#include "symtab.h"
 
 int warnlevel;
 
@@ -40,6 +36,8 @@ static int list_loc;
 static pic_instr_t *list_ptr;
 static long list_val, list_len;
 
+int default_radix;
+
 int cond_nest_count;
 int unique_id_count;
 
@@ -75,6 +73,13 @@ int org_val;
 
 int local_level;
 
+static void
+err_out(char *str)
+{
+  fputs(str, stderr);
+  fputc('\n', stderr);
+}
+
 /* Error handling */
 /*
  * Show line number/line with error message
@@ -83,20 +88,27 @@ static void
 err_line_ref(void)
 {
   struct inc_file *inc;
+  char outbuf[128];
+  int len;
 
   if(current_file != NULL) {
     inc = current_file;
     if(inc->type != INC_FILE) {
-      fprintf(stderr, "(Macro %s line %d) ",
+      sprintf(outbuf, "(Macro %-.99s line %d)",
 	      inc->v.m.sym->name, inc->linenum);
+      err_out(outbuf);
       while(inc != NULL && inc->type != INC_FILE)
 	inc = inc->next;
     }
-    fprintf(stderr, "File '%s' at line %d:\n",
+    sprintf(outbuf, "%-.99s:%d:",
 	    inc->v.f.fname, inc->linenum);
-    fputs(line_buffer, stderr);
-    if(line_buffer[0] != '\0' && line_buffer[strlen(line_buffer)-1] != '\n')
-      fputc('\n', stderr);
+    err_out(outbuf);
+    len = strlen(line_buffer);
+    if(len > 100)
+      len = 100;
+    strncpy(outbuf, line_buffer, len);
+    outbuf[len] = '\0';
+    err_out(outbuf);
   }
 }
 
@@ -106,19 +118,18 @@ err_line_ref(void)
 void
 warning(char *fmt, ...)
 {
+  char outbuf[128];
   va_list args;
 
   err_line_ref();
-  fputs("Warning: ", stderr);
+  strcpy(outbuf, "Warning: ");
   va_start(args, fmt);
-  vfprintf(stderr, fmt, args);
-  if(list_fp != NULL) {
-    fputs("Warning: ", list_fp);
-    vfprintf(list_fp, fmt, args);
-    fputc('\n', list_fp);
-  }
-  fputc('\n', stderr);
+  vsprintf(outbuf+9, fmt, args);
   va_end(args);
+  err_out(outbuf);
+  if(list_fp != NULL) {
+    fputs(outbuf, list_fp);
+  }
   warnings++;
 }
 
@@ -142,18 +153,18 @@ void
 error(int lskip, char *fmt, ...)
 {
   va_list args;
+  char outbuf[128];
 
   err_line_ref();
-  fputs("Error: ", stderr);
+  strcpy(outbuf, "Error: ");
   va_start(args, fmt);
-  vfprintf(stderr, fmt, args);
+  vsprintf(outbuf+7, fmt, args);
+  va_end(args);
+  err_out(outbuf);
   if(list_fp != NULL) {
-    fputs("Error: ", list_fp);
-    vfprintf(list_fp, fmt, args);
+    fputs(outbuf, list_fp);
     fputc('\n', list_fp);
   }
-  fputc('\n', stderr);
-  va_end(args);
   if(++errors >= MAX_ERRORS)
     fatal_error("too many errors, aborting");
 	
@@ -168,14 +179,15 @@ void
 fatal_error(char *fmt, ...)
 {
   va_list args;
+  char outbuf[128];
 
   err_line_ref();
-  fputs("Fatal error: ", stderr);
+  strcpy(outbuf, "Fatal error: ");
   va_start(args, fmt);
-  vfprintf(stderr, fmt, args);
-  fputc('\n', stderr);
+  vsprintf(outbuf+13, fmt, args);
   va_end(args);
-  exit(EXIT_FAILURE);
+  err_out(outbuf);
+  exit(EXIT_FAILURE); /* XXX possibly use longjmp() */
 }
 
 /* memory allocation */
@@ -199,7 +211,8 @@ init_assembler(void)
   pic_instr_t *cp;
   int n;
 
-  init_symtab(); /* initialize symbol table */
+  init_symtab();   /* initialize symbol table  */
+  init_includes(); /* initialize include paths */
 
   /* initialize program memory to invalid values */
   for(n = PROGMEM_MAX, cp = prog_mem; n-- > 0; *cp++ = INVALID_INSTR);
@@ -235,6 +248,8 @@ init_assembler(void)
   local_level = 0;
 
   ifskip_mode = 0;
+
+  default_radix = 10;
 }
 
 /*
@@ -265,7 +280,7 @@ void gen_code(int val)
     fatal_error("Code address out of range");
 
   if(prog_mem[prog_location] != INVALID_INSTR)
-    warning("Overlapping code at 0x%x\n", prog_location);
+    warning("Overlapping code at 0x%x", prog_location);
 
   if((list_flags & LIST_PROG) == 0) {
     list_loc = prog_location;
@@ -304,7 +319,7 @@ gen_edata(int val)
     fatal_error("Data EEPROM address out of range");
 
   if(data_eeprom[edata_location] < 0x100)
-    warning("Overlapping EEPROM data at 0x%x\n", edata_location);
+    warning("Overlapping EEPROM data at 0x%x", edata_location);
 
   if((list_flags & LIST_LOC) == 0) {
     list_loc = edata_location;
@@ -327,38 +342,38 @@ write_hex_record(FILE *fp,
 			   pic_instr_t *data, /* pointer to word data */
 			   int format) /* IHX8M or IHX16 */
 {
-  int check = 0;
+  unsigned int check = 0;
 
   switch(format) {
-    case IHX8M:
-      fprintf(fp, ":%02X%04X00", 2*reclen, 2*loc);
-      check += ((2*loc) & 0xff) + (((2*loc) >> 8) & 0xff) + 2*reclen;
-      break;
+  case IHX8M:
+    fprintf(fp, ":%02X%04X00", 2*reclen, 2*loc);
+    check += ((2*loc) & 0xff) + (((2*loc) >> 8) & 0xff) + 2*reclen;
+    break;
 
-    case IHX16:
-      fprintf(fp, ":%02X%04X00", reclen, loc);
-      check += (loc & 0xff) + ((loc >> 8) & 0xff) + reclen;
-      break;
+  case IHX16:
+    fprintf(fp, ":%02X%04X00", reclen, loc);
+    check += (loc & 0xff) + ((loc >> 8) & 0xff) + reclen;
+    break;
   }
 
   while(reclen--) {
     switch(format) {
-      case IHX8M:
-        fprintf(fp, "%02X%02X",
-		(int)(*data & 0xff), (int)((*data >> 8) & 0xff));
-	break;
+    case IHX8M:
+      fprintf(fp, "%02X%02X",
+	      (int)(*data & 0xff), (int)((*data >> 8) & 0xff));
+      break;
 
-      case IHX16:
-	fprintf(fp, "%02X%02X",
-		(int)((*data >> 8) & 0xff), (int)(*data & 0xff));
-	break;
+    case IHX16:
+      fprintf(fp, "%02X%02X",
+	      (int)((*data >> 8) & 0xff), (int)(*data & 0xff));
+      break;
     }
     check += (*data & 0xff) + ((*data >> 8) & 0xff);
     data++;
   }
 
-  /* write checksum, assumes 2-complement */
-  fprintf(fp, "%02X" HEX_EOL, (-check) & 0xff);
+  /* write checksum */
+  fprintf(fp, "%02X" HEX_EOL, (0x100 - (check & 0xff)) & 0xff);
 }
 
 /*
@@ -372,7 +387,7 @@ write_output(char *fname, int format)
   FILE *fp;
 
   if((fp = fopen(fname, "w")) == NULL)
-    fatal_error("Can't create file '%s'", fname);
+    fatal_error("Can't create file '%s': %s", fname, strerror(errno));
 
   /* program */
   for(loc = 0;;) {
@@ -394,19 +409,34 @@ write_output(char *fname, int format)
   /* PIC ID */
   if(pic_id[0] != INVALID_ID) {
     switch(pic_type->instr_set) {
-      case PIC12BIT:
-        write_hex_record(fp, 4, prog_mem_size, pic_id, format);
-	break;
+    case PIC12BIT:
+      write_hex_record(fp, 4, prog_mem_size, pic_id, format);
+      break;
 
-      case PIC14BIT:
-	write_hex_record(fp, 4, 0x2000, pic_id, format);
-	break;
+    case PIC14BIT:
+      write_hex_record(fp, 4, 0x2000, pic_id, format);
+      break;
+
+    default:
+      break;
     }
   }
 
   /* config fuses */
   if(config_fuses != INVALID_CONFIG) {
-    write_hex_record(fp, 1, pic_type->fuse_addr, &config_fuses, format);
+    switch(pic_type->instr_set) {
+    case PIC12BIT:
+      write_hex_record(fp, 1, 0xfff, &config_fuses, format);
+      break;
+    case PIC14BIT:
+      write_hex_record(fp, 1, 0x2007, &config_fuses, format);
+      break;
+    case PIC16BIT_NOMUL:
+    case PIC16BIT:
+      /* XXX Is this correct? */
+      write_hex_record(fp, 1, 0xfe00, &config_fuses, format);
+      break;
+    }
   }
 
   if(pic_type->eeprom_size > 0) { /* data EEPROM */
@@ -459,11 +489,14 @@ write_listing_line(int cond_flag)
 	if((list_flags & (LIST_PROG|LIST_EDATA|LIST_PTR)) != 0) {
 	  fputc(((list_flags & LIST_FORWARD) ? '?' : ' '), list_fp);
 	  if(list_flags & LIST_PROG) {
-	    fprintf(list_fp, "%04X ", prog_mem[list_loc]);
+	    fprintf(list_fp, "%04X ",
+		    (unsigned short)prog_mem[list_loc]);
 	  } else if(list_flags & LIST_EDATA) {
-	    fprintf(list_fp, "%04X ", data_eeprom[list_loc]);
+	    fprintf(list_fp, "%04X ",
+		    (unsigned short)data_eeprom[list_loc]);
 	  } else if(list_flags & LIST_PTR) {
-	    fprintf(list_fp, "%04X ", *list_ptr++);
+	    fprintf(list_fp, "%04X ",
+		    (unsigned short)(*list_ptr++));
 	  }
 	} else {
 	  fputs("      ", list_fp);
@@ -471,9 +504,7 @@ write_listing_line(int cond_flag)
       }
 
       fputs(line_buffer, list_fp);
-      if(line_buffer[0] != '\0'
-	 && line_buffer[strlen(line_buffer)-1] != '\n')
-	fputc('\n', list_fp);
+      fputc('\n', list_fp);
 
       list_len--;
       for(i = 0; i < list_len; i++) {
@@ -490,13 +521,19 @@ write_listing_line(int cond_flag)
 	  fputs("    ", list_fp);
 
 	if(list_flags & LIST_PROG) {
-	  fprintf(list_fp, " %04X\n", prog_mem[list_loc]);
+	  fprintf(list_fp, " %04X\n",
+		  (unsigned short)prog_mem[list_loc]);
 	} else if(list_flags & LIST_EDATA) {
-	  fprintf(list_fp, " %04X\n", data_eeprom[list_loc]);
+	  fprintf(list_fp, " %04X\n",
+		  (unsigned short)
+		  data_eeprom[list_loc]);
 	} else if(list_flags & LIST_PTR) {
-	  fprintf(list_fp, " %04X\n", *list_ptr++);
+	  fprintf(list_fp, " %04X\n",
+		  (unsigned short)(*list_ptr++));
 	}
       }
+    } else { /* if(line_buffer[0] != '\0' ... */
+      fputc('\n', list_fp);
     }
 
     if(listing_on < 0)
@@ -508,7 +545,7 @@ write_listing_line(int cond_flag)
 
 /*
  * parse and handle OPT-directive
- * (this is special as it must be done as macro definition time
+ * (this is special as it must be done at macro definition time
  *  if inside a macro)
  */
 static int
@@ -540,6 +577,9 @@ handle_opt(void)
 
 /*
  * Define a macro
+ *
+ * the check for multiple definitions of the macro name
+ * has been done before calling this function.
  */
 static void
 define_macro(char *name)
@@ -688,36 +728,37 @@ apply_patches(struct patch *patch_list)
   for(ptch = patch_list; ptch != NULL; ptch = p2) {
     p2 = ptch->next;
 
-    if(ptch->label->type == SYM_FORWARD)
+    if(ptch->label->type == SYM_FORWARD) {
       error(0, "Undefined label '%s%s'",
 	    (patch_list == global_patch_list ? "" : "="),
 	    ptch->label->name);
-    else
+    } else {
       switch(ptch->type) {
-        case PATCH8:
-	  if(pic_type->instr_set == PIC12BIT && (ptch->label->v.value & 0x100) != 0
-	     && (prog_mem[ptch->location] & 0xff00) == 0x900)
-	    error(0, "CALL address in upper half of a page (label '%s%s')",
-		  (patch_list == global_patch_list ? "" : "="),
-		  ptch->label->name);
+      case PATCH8:
+	if(pic_type->instr_set == PIC12BIT && (ptch->label->v.value & 0x100) != 0
+	   && (prog_mem[ptch->location] & 0xff00) == 0x900)
+	  error(0, "CALL address in upper half of a page (label '%s%s')",
+		(patch_list == global_patch_list ? "" : "="),
+		ptch->label->name);
 
-	  prog_mem[ptch->location] =
-	    (prog_mem[ptch->location] & 0xff00)
-	      | (ptch->label->v.value & 0xff);
-	  break;
+	prog_mem[ptch->location] =
+	  (prog_mem[ptch->location] & 0xff00)
+	  | (ptch->label->v.value & 0xff);
+	break;
 
-	case PATCH9:
-	  prog_mem[ptch->location] =
-	    (prog_mem[ptch->location] & 0xfe00)
-	      | (ptch->label->v.value & 0x1ff);
-	  break;
+      case PATCH9:
+	prog_mem[ptch->location] =
+	  (prog_mem[ptch->location] & 0xfe00)
+	  | (ptch->label->v.value & 0x1ff);
+	break;
 
-	case PATCH11:
-	  prog_mem[ptch->location] =
-	    (prog_mem[ptch->location] & 0xf800)
-	      | (ptch->label->v.value & 0x7ff);
-	  break;
+      case PATCH11:
+	prog_mem[ptch->location] =
+	  (prog_mem[ptch->location] & 0xf800)
+	  | (ptch->label->v.value & 0x7ff);
+	break;
       }
+    }
     mem_free(ptch);
   }
 }
@@ -758,8 +799,7 @@ gen_byte_c(int instr_code)
     }
   }
 
-  val = get_expression();
-  if(expr_error)
+  if(get_expression(&val) != OK)
     return FAIL;
 
   if(val < -0x80 || val > 0xff) {
@@ -803,16 +843,21 @@ assembler(char *fname)
   int op, t, symtype;
   long val;
   char *cp;
-  struct pic_type *pic;
+
+  sym = add_symbol("_PICASM", SYMTAB_GLOBAL);
+  sym->type = SYM_DEFINED;
+  sym->type2 = SYMT_CONSTANT;
+  sym->v.value = I_VERSION;
 
   if(pic_type != NULL) {
     sprintf(symname, "__%s", pic_type->name);
     sym = add_symbol(symname, SYMTAB_GLOBAL);
     sym->type = SYM_DEFINED;
+    sym->type2 = SYMT_CONSTANT;
     sym->v.value = 1;
   }
 
-  begin_include(fname);
+  begin_include(fname, 0);
   get_token();
 
   while(token_type != TOK_EOF) {
@@ -832,7 +877,7 @@ assembler(char *fname)
       sym = lookup_symbol(symname, symtype);
       if(sym != NULL && sym->type == SYM_MACRO) {
 	/* skip whitespace */
-	while(tok_char != '\n' && isspace(tok_char))
+	while(isspace(tok_char))
 	  read_src_char();
 
 	if(line_buf_ptr != NULL &&
@@ -849,94 +894,113 @@ assembler(char *fname)
 
       get_token();
       switch(token_type) {
-        case KW_MACRO:
-	  get_token();
-	  if(sym != NULL) {
-	    error(1, "Multiply defined symbol '%s%s'",
-		  (symtype == SYMTAB_LOCAL ? "=" : ""),
-		  sym->name);
-	    continue;
-	  }
-	  define_macro(symname);
-	  goto line_end;
+      case KW_MACRO:
+	get_token();
+	if(sym != NULL) {
+	  error(1, "Multiply defined symbol '%s%s'",
+		(symtype == SYMTAB_LOCAL ? "=" : ""),
+		sym->name);
+	  continue;
+	}
+	define_macro(symname);
+	goto line_end;
 
-	case KW_EQU:
-	  if(sym != NULL) {
-	    if(sym->type != SYM_FORWARD)
-	      error(0, "Multiply defined symbol '%s%s'",
-		    (symtype == SYMTAB_LOCAL ? "=" : ""),
-		    sym->name);
-	  } else
-	    sym = add_symbol(symname, symtype);
-	  get_token();
-	  sym->type = SYM_DEFINED;
-	  sym->v.value = get_expression();
-	  if(expr_error)
-	    continue; /* error_lineskip() done in expr.c */
-
-	  list_val = sym->v.value;
-	  list_flags = LIST_VAL;
-	  goto line_end;
-
-	case KW_SET:
-	  if(sym != NULL && sym->type != SYM_SET)
+      case TOK_EQUAL: /* '=' can be used instead of EQU */
+      case KW_EQU:
+	if(sym != NULL) {
+	  if(sym->type != SYM_FORWARD)
 	    error(0, "Multiply defined symbol '%s%s'",
 		  (symtype == SYMTAB_LOCAL ? "=" : ""),
 		  sym->name);
-	  else if(sym == NULL)
-	    sym = add_symbol(symname, symtype);
-	  get_token();
-	  sym->type = SYM_SET;
-	  sym->v.value = get_expression();
-	  if(expr_error)
-	    continue;
+	} else {
+	  sym = add_symbol(symname, symtype);
+	}
+	get_token();
+	sym->type = SYM_DEFINED;
+	sym->type2 = SYMT_CONSTANT;
 
-	  list_val = sym->v.value;
-	  list_flags = LIST_VAL;
-	  goto line_end;
+	if(get_expression(&sym->v.value) != OK)
+	  continue; /* error_lineskip() done in expr.c */
 
-	case TOK_COLON:
-	  get_token();
-	  goto do_label;
+	list_val = sym->v.value;
+	list_flags = LIST_VAL;
+	goto line_end;
 
-	default:
-	  if(t == 0)
-	    warning("Label not in the beginning of a line");
+      case KW_SET:
+	if(sym != NULL && sym->type != SYM_SET) {
+	  error(0, "Multiply defined symbol '%s%s'",
+		(symtype == SYMTAB_LOCAL ? "=" : ""),
+		sym->name);
+	} else if(sym == NULL) {
+	  sym = add_symbol(symname, symtype);
+	}
+	get_token();
+	sym->type = SYM_SET;
+	sym->type2 = SYMT_CONSTANT;
+
+	if(get_expression(&sym->v.value) != OK)
+	  continue;
+
+	list_val = sym->v.value;
+	list_flags = LIST_VAL;
+	goto line_end;
+
+      case TOK_COLON:
+	get_token();
+	goto do_label;
+
+      default:
+	if(t == 0)
+	  warning("Label not in the beginning of a line");
 
 do_label:
-	  switch(O_Mode) {
-	    case O_PROGRAM:
-	      t = prog_location;
-	      break;
-
-	    case O_REGFILE:
-	      t = reg_location;
-	      break;
-
-	    case O_EDATA:
-	      t = edata_location;
-	      break;
-
-	    case O_NONE:
-	      t = org_val;
-	      break;
-	  }
-
-	  if(t < 0) {
-	    error(0, "ORG value not set");
-	  } else {
-	    if(sym != NULL && sym->type != SYM_FORWARD)
-	      error(0, "Multiply defined symbol '%s%s'",
-		    (symtype == SYMTAB_LOCAL ? "=" : ""),
-		    sym->name);
-	    if(sym == NULL)
-	      sym = add_symbol(symname, symtype);
-	    sym->type = SYM_DEFINED;
-	    sym->v.value = t;
-	    list_loc = t;
-	    list_flags = LIST_LOC;
-	  }
+	switch(O_Mode) {
+	case O_PROGRAM:
+	  t = prog_location;
 	  break;
+
+	case O_REGFILE:
+	  t = reg_location;
+	  break;
+
+	case O_EDATA:
+	  t = edata_location;
+	  break;
+
+	case O_NONE:
+	  t = org_val;
+	  break;
+	}
+
+	if(t < 0) {
+	  error(0, "ORG value not set");
+	} else {
+	  if(sym != NULL && sym->type != SYM_FORWARD) {
+	    error(0, "Multiply defined symbol '%s%s'",
+		  (symtype == SYMTAB_LOCAL ? "=" : ""),
+		  sym->name);
+	  }
+	  if(sym == NULL) {
+	    sym = add_symbol(symname, symtype);
+	  }
+	  sym->type = SYM_DEFINED;
+	  switch(O_Mode) {
+	  case O_PROGRAM:
+	    sym->type2 = SYMT_LABEL;
+	    break;
+
+	  case O_REGFILE:
+	    sym->type2 = SYMT_REGFILE;
+	    break;
+
+	  default:
+	    break;
+	  }
+	  sym->v.value = t;
+	  list_loc = t;
+	  list_flags = LIST_LOC;
+	}
+	break;
       }
     }
 
@@ -969,7 +1033,7 @@ do_label:
     get_token();
 
     switch(op) {
-      case KW_INCLUDE:
+    case KW_INCLUDE:
       if(token_type != TOK_STRCONST) {
 	error(1, "Missing file name after INCLUDE");
 	continue;
@@ -980,12 +1044,13 @@ do_label:
       if(token_type != TOK_NEWLINE && token_type != TOK_EOF)
 	error(0, "Extraneous characters after a valid source line");
 
-      begin_include(symname);
+      begin_include(symname, 1);
 
       write_listing_line(0);
       get_token();
       continue;
 
+    case TOK_EQUAL:
     case KW_SET:
     case KW_EQU:
       if(sym == NULL)
@@ -1041,7 +1106,7 @@ do_label:
       if(sym != NULL)
 	error(0, "Label not allowed with IF");
 
-      val = get_expression();
+      get_expression(&val); /*note: error status ignored*/
 
       if(token_type != TOK_NEWLINE && token_type != TOK_EOF)
 	error(0, "Extraneous characters after a valid source line");
@@ -1127,8 +1192,7 @@ do_label:
 	break;
       }
 
-      val = get_expression();
-      if(expr_error)
+      if(get_expression(&val) != OK)
 	continue;
 
       if(val < 0 || val >= prog_mem_size) {
@@ -1170,8 +1234,7 @@ do_label:
       break;
 
     case KW_DS:
-      val = get_expression();
-      if(expr_error)
+      if(get_expression(&val) != OK)
 	continue;
 
       if(O_Mode == O_NONE) {
@@ -1179,8 +1242,9 @@ do_label:
 	if(org_val < 0) {
 	  error(0, "ORG value not set");
 	  reg_location = 0;
-	} else
+	} else {
 	  reg_location = org_val;
+	}
       }
 
       if(O_Mode != O_REGFILE)
@@ -1192,6 +1256,78 @@ do_label:
 	list_flags = LIST_LOC;
 	reg_location += val;
       }
+
+      if(sym != NULL && sym->type2 == SYMT_UNKNOWN)
+	sym->type2 = SYMT_REGFILE;
+      break;
+
+    case KW_CBLOCK:
+      if(sym != NULL)
+	error(0, "Label not allowed with CBLOCK");
+
+      if(token_type != TOK_NEWLINE && token_type != TOK_EOF) {
+	if(get_expression(&val) != OK) {
+	  error_lineskip();
+	} else {
+	  reg_location = val;
+	  if(token_type != TOK_NEWLINE && token_type != TOK_EOF)
+	    error(0, "Extraneous characters after a valid source line");
+	  skip_eol();
+	}
+      }
+
+      if(reg_location < 0) {
+	error(0, "ORG/CBLOCK register address not set");
+	reg_location = 0;
+      }
+
+      /* not really implemented .... */
+      while(token_type != KW_ENDC) {
+	list_loc = -1;
+	while(token_type != TOK_NEWLINE) {
+	  if(token_type == TOK_EOF) {
+	    error(0, "Unexpected EOF inside a CBLOCK");
+	    break;
+	  }
+
+	  if(token_type != TOK_IDENTIFIER) {
+	    error(1, "identifier expected");
+	    goto out;
+	  }
+
+	  if(reg_location >= reg_file_limit)
+	    fatal_error("Register file address out of range");
+
+	  if(list_loc < 0)
+	    list_loc = reg_location;
+
+	  sym = lookup_symbol(token_string, SYMTAB_GLOBAL);
+	  if(sym != NULL) {
+	    if(sym->type != SYM_FORWARD) {
+	      error(0, "Symbol type conflict (%s)", token_string); 
+	      sym = NULL;
+	    }
+	  } else {
+	    sym = add_symbol(token_string, SYMTAB_GLOBAL);
+	  }
+
+	  if(sym != NULL) {
+	    sym->type = SYM_DEFINED;
+	    sym->type2 = SYMT_REGFILE;
+	    sym->v.value = reg_location;
+	    reg_location++;
+	  }
+
+	  get_token();
+	  if(token_type == TOK_COMMA)
+	    get_token();
+	}
+	list_flags = (list_loc >= 0 ? LIST_LOC : 0);
+	write_listing_line(0);
+	get_token(); /* skip the newline */
+      }
+      get_token(); /* skip KW_ENDC token */
+out:
       break;
 
     case KW_EDATA:
@@ -1209,8 +1345,7 @@ do_label:
 	    gen_edata((int)((unsigned char)(*cp)));
 	  get_token();
 	} else {
-	  val = get_expression();
-	  if(expr_error)
+	  if(get_expression(&val) != OK)
 	    break;
 
 	  if(val < 0 || val > 0xff) {
@@ -1229,11 +1364,6 @@ do_label:
       if(pic_type == NULL)
 	fatal_error("PIC device type not set");
 
-      if(config_fuses != INVALID_CONFIG) {
-	error(1, "Multiple CONFIG definitions");
-	continue;
-      }
-
       parse_config();
 
       list_flags = LIST_PTR;
@@ -1248,32 +1378,38 @@ do_label:
 	continue;
       }
 
+
       cp = token_string;
       if(strncasecmp(token_string, "PIC", 3) == 0)
 	cp += 3;
 
-      for(pic = pic_types; pic->name != NULL; pic++) {
-	if(strcasecmp(pic->name, cp) == 0)
+      {
+	struct pic_type *pic;
+
+	for(pic = pic_types; pic->name != NULL; pic++) {
+	  if(strcasecmp(pic->name, cp) == 0)
+	    break;
+	}
+	if(pic->name == NULL)
+	  fatal_error("Invalid PIC device type");
+
+	get_token();
+	if(pic_type == pic)
 	  break;
-      }
-      if(pic->name == NULL)
-	fatal_error("Invalid PIC device type");
 
-      get_token();
-      if(pic_type == pic)
-	break;
-
-      if(pic_type != NULL) {
-	error(1, "Duplicate DEVICE setting");
-	continue;
+	if(pic_type != NULL) {
+	  error(1, "Duplicate DEVICE setting");
+	  continue;
+	}
+	pic_type = pic;
+	prog_mem_size = pic_type->progmem_size;
+	reg_file_limit = pic_type->regfile_limit;
       }
-      pic_type = pic;
-      prog_mem_size = pic_type->progmem_size;
-      reg_file_limit = pic_type->regfile_limit;
 
       sprintf(symname, "__%s", pic_type->name);
       sym = add_symbol(symname, SYMTAB_GLOBAL);
       sym->type = SYM_DEFINED;
+      sym->type2 = SYMT_CONSTANT;
       sym->v.value = 1;
       break;
 
@@ -1292,11 +1428,18 @@ do_label:
        * Microchip 1994 databook specfically says that ID is not
        * implemented on PIC16C6x/74... but 1995/96 databook
        * has different information about that...
+       *
+       * But seems that the 16-bit PICs don't have an id
+       *
        */
 
+      if(pic_type->instr_set == PIC16BIT) {
+	error(1, "ID locations not supported on PIC%s", pic_type->name);
+	continue;
+      }
+
       for(t = 0;;) {
-	val = get_expression();
-	if(expr_error)
+	if(get_expression(&val) != OK)
 	  continue;
 
 	if(val < 0 || val > 0x3fff)
@@ -1330,15 +1473,23 @@ do_label:
       /* mnemonics */
 
     default:
+      if(pic_type == NULL)
+	fatal_error("PIC device type not set");
+
       switch(pic_type->instr_set) {
         case PIC12BIT:
 	  t = assemble_12bit_mnemonic(op);
 	  break;
 
 	case PIC14BIT:
-	default:
+        default:
 	  t = assemble_14bit_mnemonic(op);
 	  break;
+
+        case PIC16BIT_NOMUL:
+        case PIC16BIT:
+	  t = assemble_16bit_mnemonic(op);
+	  break;
 	}
       if(t != OK)
 	continue;
@@ -1372,15 +1523,23 @@ line_end2:
   apply_patches(global_patch_list);
 }
 
+static void
+usage(void)
+{
+  fputs("Usage: picasm [-o<objname>] [-l<listfile>] [-s] [-ihx8m/ihx16] [-i<incdir>]\n"
+	"              [-pic<device>] [-d<sym>[=<value>]] [-w[n]] <filename>\n", stderr);
+  exit(EXIT_FAILURE);
+}
+
 /*
  * main program
  */
 int
 main(int argc, char *argv[])
 {
-  static char in_filename[256], out_filename[256], list_filename[256];
+  static char in_filename[256], out_filename[256], list_filename[256], sym_filename[256];
   static int out_format = IHX8M;
-  static int listing = 0, symdump = 0;
+  int listing = 0, symdump = 0;
   char *p;
   time_t ti;
   struct tm *tm;
@@ -1388,84 +1547,143 @@ main(int argc, char *argv[])
   pic_type = NULL;
   out_filename[0] = '\0';
   list_filename[0] = '\0';
+  sym_filename[0] = '\0';
   warnlevel = 0;
 
+  init_assembler();
+
   while(argc > 1 && argv[1][0] == '-') {
     switch(argv[1][1]) {
-      case 'o': /* output file name */
-        if(argv[1][2] != '\0')
-	  strcpy(out_filename, &argv[1][2]);
-	else {
-	  if(argc < 3) {
-	    fputs("-o option requires a file name\n", stderr);
-	    exit(EXIT_FAILURE);
-	  }
-	  strcpy(out_filename, argv[2]);
+    case 'o': /* output file name */
+      if(argv[1][2] != '\0') {
+	strcpy(out_filename, &argv[1][2]);
+      } else {
+	if(argc < 3) {
+	  fputs("-o option requires a file name\n", stderr);
+	  exit(EXIT_FAILURE);
+	}
+	strcpy(out_filename, argv[2]);
+	argc--;
+	argv++;
+      }
+      break;
+
+    case 'i': case 'I': /* output hex format (ihx8m/ihx16) */
+      if(strcasecmp(&argv[1][1], "ihx8m") == 0) {
+	out_format = IHX8M;
+      } else if(strcasecmp(&argv[1][1], "ihx16") == 0) {
+	out_format = IHX16;
+      } else {
+	if(argv[1][2] != '\0') {
+	  add_include_path(&argv[1][2]);
+	} else {
+	  if(argc < 3)
+	    usage();
+	  add_include_path(argv[2]);
 	  argc--;
 	  argv++;
 	}
-	break;
+      }
+      break;
 
-      case 'i': case 'I': /* output hex format (ihx8m/ihx16) */
-	if(strcasecmp(&argv[1][1], "ihx8m") == 0)
-	  out_format = IHX8M;
-	else if(strcasecmp(&argv[1][1], "ihx16") == 0)
-	  out_format = IHX16;
-	else
-	  goto usage;
-	break;
+    case 'd': /* define symbols */
+      {
+	char symname[256], *p;
+	long val = 1;
+	struct symbol *sym;
 
-      case 'p': case 'P': /* PIC type */
-	if(!((argv[1][2] == 'i' || argv[1][2] == 'I') &&
-	     (argv[1][3] == 'c' || argv[1][3] == 'C')))
-	  goto usage;
+	p = &argv[1][2];
+	if(!isalpha((unsigned char)(*p)) && *p != '\0') {
+badsym:
+	  fputs("Bad symbol name (-d option)\n", stderr);
+	  exit(EXIT_FAILURE);
+	}
+	while(*p != '\0' && *p != '=') {
+	  if(!isalnum((unsigned char)(*p)) && *p != '_')
+	    goto badsym;
+	  if(p >= &argv[1][2] + 255) {
+	    fputs("Symbol too long (-d option)\n", stderr);
+	    exit(EXIT_FAILURE);
+	  }
+	  p++;
+	}
+	if(p == &argv[1][2])
+	  goto badsym;
+	memcpy(symname, &argv[1][2], p - &argv[1][2]);
+	symname[p - &argv[1][2]] = '\0';
+	if(*p == '=') {
+	  char *ep;
 
-	for(pic_type = pic_types; pic_type->name != NULL; pic_type++) {
-	  if(strcasecmp(pic_type->name, &argv[1][4]) == 0)
-	    break;
+	  val = strtol(p+1, &ep, 0);
+	  if(*ep != '\0') {
+	    fputs("Invalid numeric value (-d option)\n", stderr);
+	    exit(EXIT_FAILURE);
+	  }
 	}
 
-	if(pic_type->name == NULL) {
-	  fprintf(stderr, "Invalid device type '%s'\n", &argv[1][1]);
+	if(lookup_symbol(symname, SYMTAB_GLOBAL) != NULL) {
+	  fprintf(stderr, "Multiply defined symbol '%s'\n", symname);
 	  exit(EXIT_FAILURE);
 	}
 
-	prog_mem_size = pic_type->progmem_size;
-	reg_file_limit = pic_type->regfile_limit;
-	break;
+	sym = add_symbol(symname, SYMTAB_GLOBAL);
+	sym->type = SYM_DEFINED;
+	sym->type2 = SYMT_CONSTANT;
+	sym->v.value = val;
+      }
+      break;
 
-      case 'l': /* listing/list filename */
-	listing = 1;
-	if(argv[1][2] != '\0')
-	  strcpy(list_filename, &argv[1][2]);
-	break;
+    case 'p': case 'P': /* PIC type */
+      if(!((argv[1][2] == 'i' || argv[1][2] == 'I') &&
+	   (argv[1][3] == 'c' || argv[1][3] == 'C')))
+	usage();
 
-      case 's':
-        symdump = 1;
-        break;
+      for(pic_type = pic_types; pic_type->name != NULL; pic_type++) {
+	if(strcasecmp(pic_type->name, &argv[1][4]) == 0)
+	  break;
+      }
 
-      case 'w': /* warning mode (gives some more warnings) */
-	if(argv[1][2] != '\0') {
-	  warnlevel = atoi(&argv[1][2]);
-	} else {
-	  warnlevel = 1;
-	}
-	break;
+      if(pic_type->name == NULL) {
+	fprintf(stderr, "Invalid device type '%s'\n", &argv[1][1]);
+	exit(EXIT_FAILURE);
+      }
 
-      case 'v': /* version info */
-	fprintf(stderr,
-		"12/14-bit PIC assembler " VERSION
-		" -- Copyright 1995-1998 by Timo Rossi\n");
-	break;
+      prog_mem_size = pic_type->progmem_size;
+      reg_file_limit = pic_type->regfile_limit;
+      break;
+
+    case 'l': /* listing/list filename */
+      listing = 1;
+      if(argv[1][2] != '\0')
+	strcpy(list_filename, &argv[1][2]);
+      break;
+
+    case 's':
+      symdump = 1;
+      break;
+
+    case 'w': /* warning mode (gives some more warnings) */
+      if(argv[1][2] != '\0') {
+	warnlevel = atoi(&argv[1][2]);
+      } else {
+	warnlevel = 1;
+      }
+      break;
+
+    case 'v': /* version info */
+      fprintf(stderr,
+	      "12/14/16-bit PIC assembler " VERSION
+	      " -- Copyright 1995-2000 Timo Rossi\n");
+      break;
 	
-      case '-': /* end of option list */
-      case '\0':
-	argc--;
-	argv++;
-	goto opt_done;
+    case '-': /* end of option list */
+    case '\0':
+      argc--;
+      argv++;
+      goto opt_done;
 
-      default:
-	goto usage;
+    default:
+      usage();
     }
     argc--;
     argv++;
@@ -1473,10 +1691,7 @@ main(int argc, char *argv[])
 
 opt_done:
   if(argc != 2) {
-usage:
-    fputs("Usage: picasm [-o<objname>] [-l<listfile>] [-s] [-ihx8m/ihx16]\n"
-	  "              [-pic<device>] [-w[n]] <filename>\n", stderr);
-    exit(EXIT_FAILURE);
+    usage();
   }
 
   strncpy(in_filename, argv[1], sizeof(in_filename)-1);
@@ -1491,8 +1706,6 @@ usage:
   if(strchr(out_filename, '.') == NULL)
     strcat(out_filename, ".hex");
 
-  init_assembler();
-
   list_fp = NULL;
   if(listing) {
     if(list_filename[0] == '\0') {
@@ -1503,7 +1716,7 @@ usage:
     }
 
     if((list_fp = fopen(list_filename, "w")) == NULL)
-      fatal_error("Can't create listing file '%s'", list_filename);
+      fatal_error("Can't create listing file '%s': %s", list_filename, strerror(errno));
 
     ti = time(NULL);
     tm = localtime(&ti);
@@ -1515,7 +1728,7 @@ usage:
 
   assembler(in_filename);
   if(errors == 0) {
-    if(code_generated)
+    if(pic_type != NULL && code_generated)
       write_output(out_filename, out_format);
     else
       fputs("No code generated\n", stderr);
@@ -1532,5 +1745,6 @@ usage:
         dump_symtab(list_fp);
       fclose(list_fp);
     }
+
   return EXIT_SUCCESS;
 }
