@@ -1,18 +1,22 @@
 /*
  * picasm -- token.c
  *
+ * Copyright 1995-2000 Timo Rossi, <trossi@iki.fi>
+ * See the file LICENSE for license terms.
+ *
  * Include handling, macro expansion, lexical analysis
  *
- * Timo Rossi <trossi@iki.fi>
- * 
  */
 
 #include <stdio.h>
 #include <string.h>
 #include <stdlib.h>
 #include <ctype.h>
+#include <errno.h>
 
 #include "picasm.h"
+#include "token.h"
+#include "symtab.h"
 
 /*
  * keyword table for tokenizer
@@ -33,6 +37,9 @@ static char Keyword_Table[] = {
   "org\0"
   "ds\0"
   "edata\0"
+  "dt\0"
+  "cblock\0"
+  "endc\0"
   "config\0"
   "picid\0"
   "device\0"
@@ -84,6 +91,35 @@ static char Keyword_Table[] = {
   "xorlw\0"
   "xorwf\0"
 
+/* 16-bit PIC instruction mnemonics */
+  "addwfc\0"
+  "btg\0"
+  "cpfseq\0"
+  "cpfsgt\0"
+  "cpfslt\0"
+  "daw\0"
+  "dcfsnz\0"
+  "infsnz\0"
+  "lcall\0"
+  "movfp\0"
+  "movpf\0"
+  "movlb\0"
+  "movlr\0"
+  "mullw\0"
+  "mulwf\0"
+  "negw\0"
+  "rlcf\0"
+  "rlncf\0"
+  "rrcf\0"
+  "rrncf\0"
+  "setf\0"
+  "subwfb\0"
+  "tablrd\0"
+  "tablwt\0"
+  "tlrd\0"
+  "tlwt\0"
+  "tstfsz\0"
+
   "\0"
 };
 
@@ -96,34 +132,108 @@ long token_int_val;
 
 int ifskip_mode; /* TRUE when skipping code inside if..endif */
 
+struct incpath_entry {
+  struct incpath_entry *next;
+  char dirname[1]; /*struct hack*/
+};
+
+static struct incpath_entry *include_path, *inc_path_next;
+
+void
+init_includes(void)
+{
+  include_path = NULL;
+  inc_path_next = NULL;
+}
+
 /*
  * include file handling
  */
 void
-begin_include(char *fname)
+add_include_path(char *dirname)
+{
+  int addslash = 0;
+  char *p;
+  struct incpath_entry *ent;
+
+  if(dirname == NULL || *dirname == '\0')
+    fatal_error("Empty directory path");
+
+  p = dirname + (strlen(dirname) - 1);
+#if defined(MSDOS) || defined(_WIN32) || defined(AMIGA)
+#ifdef AMIGA
+  if(!(*p == ':' || *p == '/'))
+    addslash = 1;
+#else
+  if(!(*p == ':' || *p == '/' || *p == '\\'))
+    addslash = 1;
+#endif
+#else /* defined(MSDOS) ... */
+  /* Unix etc. */
+  if(!(*p == '/'))
+    addslash = 1;
+#endif
+
+  ent = mem_alloc(sizeof(struct incpath_entry) + strlen(dirname) + addslash);
+  ent->next = NULL;
+  strcpy(ent->dirname, dirname);
+  if(addslash)
+    strcat(ent->dirname, "/");
+
+  if(include_path == NULL) {
+    include_path = ent;
+    inc_path_next = ent;
+  } else {
+    inc_path_next->next = ent;
+    inc_path_next = ent;
+  }
+}
+
+void
+begin_include(char *fname, int dopath)
 {
   struct inc_file *p;
+  struct incpath_entry *ent;
+  FILE *fp;
+  char fnamebuf[512];
+  int err;
+
+  strcpy(fnamebuf, fname);
+  fp = fopen(fnamebuf, "r");
+  err = errno;
+  if(fp == NULL && dopath) {
+    for(ent = include_path; ent != NULL; ent = ent->next) {
+      int elen = strlen(ent->dirname);
+      if(elen + strlen(fname) >= sizeof(fnamebuf)-1) {
+	warning("Include directory/file name too long");
+	continue;
+      }
+      strcpy(fnamebuf, ent->dirname);
+      strcpy(fnamebuf+elen, fname);
+      if((fp = fopen(fnamebuf, "r")) != NULL)
+	break;
+      err = errno;
+    } /*for*/
+  } /*if*/
+
+  if(fp == NULL) {
+    if(current_file == NULL) {
+      fatal_error("Can't open '%s': %s", fnamebuf, strerror(err));
+    } else {
+      error(0, "Can't open include file '%s': %s", fnamebuf, strerror(err));
+    }
+    line_buf_ptr = NULL;
+    tok_char = ' ';
+    return;
+  }
 
   p = mem_alloc(sizeof(struct inc_file));
   p->type = INC_FILE;
-  p->v.f.fname = mem_alloc(strlen(fname)+1);
-  strcpy(p->v.f.fname, fname);
+  p->v.f.fname = mem_alloc(strlen(fnamebuf)+1);
+  strcpy(p->v.f.fname, fnamebuf);
   p->linenum = 0;
   p->cond_nest_count = cond_nest_count;
-
-  if((p->v.f.fp = fopen(p->v.f.fname, "r")) == NULL) {
-    if(current_file == NULL) {
-      fatal_error("Can't open '%s'", p->v.f.fname);
-    } else {
-      error(0, "Can't open include file '%s'", p->v.f.fname);
-      free(p->v.f.fname);
-      free(p);
-      line_buf_ptr = NULL;
-      tok_char = ' ';
-      return;
-    }
-  }
-
+  p->v.f.fp = fp;
   p->next = current_file;
   current_file = p;
   line_buf_ptr = NULL;
@@ -148,16 +258,16 @@ end_include(void)
     p = current_file->next;
     if(current_file->type == INC_FILE) {
       fclose(current_file->v.f.fp);
-      free(current_file->v.f.fname);
+      mem_free(current_file->v.f.fname);
     } else { /* free macro arguments */
       arg1 = current_file->v.m.args;
       while(arg1 != NULL) {
 	arg2 = arg1->next;
-	free(arg1);
+	mem_free(arg1);
 	arg1 = arg2;
       }
     }
-    free(current_file);
+    mem_free(current_file);
     current_file = p;
   }
 }
@@ -173,6 +283,7 @@ expand_macro(struct symbol *sym)
   char *cp;
   int narg;
   int parcnt, d_char;
+  int nch;
 
   write_listing_line(0); /* list the macro call line */
 
@@ -187,14 +298,14 @@ expand_macro(struct symbol *sym)
   arg = NULL;
 
   for(narg = 1;;narg++) {
-    while(tok_char != '\n' && isspace(tok_char)) /* skip whitespace */
+    while(isspace(tok_char)) /* skip whitespace */
       read_src_char();
 
-    if(tok_char == '\n' || tok_char == '\0' ||
-       tok_char == ';' || tok_char == EOF)
+    if(tok_char == '\0' || tok_char == ';' || tok_char == EOF)
       break;
 
     cp = line_buf_ptr-1;
+    nch = 0;
 
     /*
      * Macro parameters are separated by commas. However, strings and
@@ -207,8 +318,7 @@ expand_macro(struct symbol *sym)
 
     parcnt = 0; /* parenthesis nesting count */
 
-    while(!isspace(tok_char) &&
-	  tok_char != '\n' && tok_char != '\0' &&
+    while(!isspace(tok_char) && tok_char != '\0' &&
 	  tok_char != ';' && tok_char != EOF) {
       if(parcnt == 0 && tok_char == ',')
 	break;
@@ -223,35 +333,33 @@ expand_macro(struct symbol *sym)
 
 	do {
 	  read_src_char();
-	}
-	while(tok_char != d_char && tok_char != '\n' &&
-	      tok_char != '\0' && tok_char != EOF);
+	  nch++;
+	} while(tok_char != d_char && tok_char != '\0' && tok_char != EOF);
 
 	if(tok_char != d_char)
 	  break;
       }
 
       read_src_char();
+      nch++;
     }
 
     if(narg >= 10)
       warning("Too many macro arguments (max. 9)");
 
     if(arg == NULL) {
-      arg = mem_alloc(sizeof(struct macro_arg)
-		      +(line_buf_ptr-cp-1));
+      arg = mem_alloc(sizeof(struct macro_arg) + nch);
       minc->v.m.args = arg;
     } else {
-      arg->next = mem_alloc(sizeof(struct macro_arg)
-			    +(line_buf_ptr-cp-1));
+      arg->next = mem_alloc(sizeof(struct macro_arg) + nch);
       arg = arg->next;
     }
-    strncpy(arg->text, cp, line_buf_ptr-cp-1);
-    arg->text[line_buf_ptr-cp-1] = '\0';
+    strncpy(arg->text, cp, nch);
+    arg->text[nch] = '\0';
     arg->next = NULL;
 
     /* skip whitespace */
-    while(tok_char != '\n' && isspace(tok_char))
+    while(isspace(tok_char))
       read_src_char();
     if(tok_char != ',')
       break;
@@ -259,8 +367,7 @@ expand_macro(struct symbol *sym)
     read_src_char();
   }
 
-  if(tok_char != ';' && tok_char != '\n' &&
-     tok_char != '\0' && tok_char != EOF)
+  if(tok_char != ';' && tok_char != '\0' && tok_char != EOF)
     error(0, "Extraneous characters after a valid source line");
 
   minc->next = current_file;
@@ -283,80 +390,94 @@ read_src_char(void)
   struct macro_arg *arg;
   static char tmpbuf[12];
 
-  if(line_buf_ptr == NULL || *line_buf_ptr == '\0') {
-    if(current_file == NULL) {
-      tok_char = EOF;
-      return;
-    }
+  if(line_buf_ptr != NULL) {
+    tok_char = (unsigned char)(*line_buf_ptr++);
+    if(tok_char == '\0')
+      line_buf_ptr = NULL;
+    return;
+  }
+
+  if(current_file == NULL) {
+    tok_char = EOF;
+    return;
+  }
 
 getc1:
-    if(current_file->type == INC_MACRO)	{
-      if(current_file->v.m.ml == NULL) {
+  if(current_file->type == INC_MACRO)	{
+    if(current_file->v.m.ml == NULL) {
+      end_include();
+      goto getc1;
+    }
+
+    scp = current_file->v.m.ml->text;
+    dcp = line_buffer;
+    while(*scp != '\0' && dcp < &line_buffer[sizeof(line_buffer)]) {
+      if(*scp == '\\') {
+	scp++;
+	if(*scp >= '1' && *scp <= '9') { /* macro arg */
+	  parm = *scp - '1'; /* macro arg #, starting from 0 */
+	  for(arg = current_file->v.m.args;
+	      arg != NULL && parm > 0; arg = arg->next, parm--);
+	  if(arg != NULL) {
+	    for(pcp = arg->text; *pcp != '\0' &&
+		  dcp < &line_buffer[sizeof(line_buffer)];)
+	      *dcp++ = *pcp++;
+	  }
+	  scp++;
+	} else if(*scp == '0' || *scp == '@') {
+	  sprintf(tmpbuf, "%03d", current_file->v.m.uniq_id);
+
+	  for(pcp = tmpbuf; *pcp != '\0' &&
+		dcp < &line_buffer[sizeof(line_buffer)];)
+	    *dcp++ = *pcp++;
+	  
+	  scp++;
+	} else if(*scp == '#') { /* number of arguments */
+	  for(parm = 0, arg = current_file->v.m.args;
+	      arg != NULL; arg = arg->next, parm++);
+	  
+	  sprintf(tmpbuf, "%d", parm);
+	  
+	  for(pcp = tmpbuf; *pcp != '\0' &&
+		dcp < &line_buffer[sizeof(line_buffer)];)
+	    *dcp++ = *pcp++;
+
+	  scp++;
+	} else
+	  *dcp++ = *scp;
+      } else
+	*dcp++ = *scp++;
+    }
+    if(dcp == &line_buffer[sizeof(line_buffer)]) {
+      error(0, "Line buffer overflow");
+      dcp--;
+    }
+
+    *dcp = '\0'; /* NUL-terminate the line */
+    current_file->v.m.ml = current_file->v.m.ml->next;
+  } else {
+    if(fgets(line_buffer, sizeof(line_buffer)-1,
+	     current_file->v.f.fp) == NULL) {
+      if(current_file->next != NULL) {
 	end_include();
 	goto getc1;
       }
-
-      scp = current_file->v.m.ml->text;
-      dcp = line_buffer;
-      while(*scp != '\0' && dcp < &line_buffer[sizeof(line_buffer)]) {
-	if(*scp == '\\') {
-	  scp++;
-	  if(*scp >= '1' && *scp <= '9') { /* macro arg */
-	    parm = *scp - '1'; /* macro arg #, starting from 0 */
-	    for(arg = current_file->v.m.args;
-		arg != NULL && parm > 0; arg = arg->next, parm--);
-	    if(arg != NULL) {
-	      for(pcp = arg->text; *pcp != '\0' &&
-		  dcp < &line_buffer[sizeof(line_buffer)];)
-		*dcp++ = *pcp++;
-	    }
-	    scp++;
-	  } else if(*scp == '0' || *scp == '@') {
-	    sprintf(tmpbuf, "%03d", current_file->v.m.uniq_id);
-
-	    for(pcp = tmpbuf; *pcp != '\0' &&
-		dcp < &line_buffer[sizeof(line_buffer)];)
-	      *dcp++ = *pcp++;
-
-	    scp++;
-	  } else if(*scp == '#') { /* number of arguments */
-	    for(parm = 0, arg = current_file->v.m.args;
-		arg != NULL; arg = arg->next, parm++);
-
-	    sprintf(tmpbuf, "%d", parm);
-
-	    for(pcp = tmpbuf; *pcp != '\0' &&
-		dcp < &line_buffer[sizeof(line_buffer)];)
-	      *dcp++ = *pcp++;
-
-	    scp++;
-	  } else
-	    *dcp++ = *scp;
-	} else
-	  *dcp++ = *scp++;
-      }
-      if(dcp == &line_buffer[sizeof(line_buffer)]) {
-	error(0, "Line buffer overflow");
-	dcp--;
-      }
-
-      *dcp = '\0'; /* NUL-terminate the line */
-      current_file->v.m.ml = current_file->v.m.ml->next;
-    } else {
-      if(fgets(line_buffer, sizeof(line_buffer)-1,
-	       current_file->v.f.fp) == NULL) {
-	if(current_file->next != NULL) {
-	  end_include();
-	  goto getc1;
-	}
-	tok_char = EOF;
-	return;
-      }
+      tok_char = EOF;
+      return;
+    }
+    if(line_buffer[0] != '\0') {
+      char *p;
+      
+      p = line_buffer + strlen(line_buffer) - 1;
+      if(*p == '\n')
+	*p = '\0';
     }
-    current_file->linenum++;
-    line_buf_ptr = line_buffer;
   }
+  current_file->linenum++;
+  line_buf_ptr = line_buffer;
   tok_char = ((unsigned char)(*line_buf_ptr++));
+  if(tok_char == '\0')
+    line_buf_ptr = NULL;
 }
 
 /*
@@ -369,26 +490,24 @@ get_token(void)
   int tp, base;
   char *cp;
 
-  for(;;) {
-    /*
-     * skip spaces
-     */
-    while(tok_char != '\n' && isspace(tok_char))
-      read_src_char();
+  /*
+   * skip spaces
+   */
+  while(isspace(tok_char))
+    read_src_char();
 
-    if(tok_char == EOF)	{
-      token_type = TOK_EOF;
-      token_string[0] = '\0';
-      return;
-    }
-
-    if(tok_char != ';')
-      break;
+  if(tok_char == EOF)	{
+    token_type = TOK_EOF;
+    token_string[0] = '\0';
+    return;
+  }
 
+  if(tok_char == ';') {
     /* comment */
-    line_buf_ptr = NULL;
-    tok_char = '\n';
-
+    skip_eol();
+    token_type = TOK_NEWLINE;
+    strcpy(token_string, "\\n");
+    return;
   } /* for(;;) */
 
 /*
@@ -451,7 +570,7 @@ get_token(void)
       read_src_char();
     }
 
-    base = 10;
+    base = default_radix;
     switch(tok_char) {
       case 'H': /* hex */
       case 'h':
@@ -588,7 +707,6 @@ get_token(void)
  * non-numeric & non-alpha tokens
  */
   switch(tok_char) {
-    case '\n':
     case '\0':
       token_type = TOK_NEWLINE;
       strcpy(token_string, "\\n");
@@ -728,11 +846,27 @@ get_token(void)
 	} while(tp < TOKSIZE-1 && isxdigit(tok_char));
 
       token_string[tp] = '\0';
-      token_int_val = strtoul(&token_string[1], NULL, 16);
+      token_int_val = strtoul(&token_string[0], NULL, 16);
       token_type = TOK_INTCONST;
       /* should put range check here */
       return;
 
+      /*
+       * "#include" and "include" both generate KW_INCLUDE
+       */
+    case '#':
+      if(strncasecmp(line_buf_ptr, "include", 7) != 0)
+	goto invalid_token;
+      strcpy(token_string, "#include");
+      line_buf_ptr += 7;
+      tok_char = (unsigned char)(*line_buf_ptr);
+      if(*line_buf_ptr == '\0') {
+	line_buf_ptr = NULL;
+	tok_char = ' ';
+      }
+      token_type = KW_INCLUDE;
+      return;
+
     case '\\':
       token_type = TOK_BACKSLASH;
       break;
